#include <atomic>
#include <iostream>
#include <mutex>
#include <windows.h>
#include <vector>
#include <fstream>
#include <thread>
#include <tlhelp32.h>
std::atomic<bool> pauseRequested(false);
std::atomic<bool> resumeRequested(false);
std::mutex pauseMutex;
HANDLE outfile=CreateFile("out",GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);

void listenForPauseResume() {
    while (true) {
        std::cin.get();
        std::lock_guard<std::mutex> lock(pauseMutex);
        if (!pauseRequested.load() && !resumeRequested.load())
            pauseRequested.store(true);
        else if (pauseRequested.load() && !resumeRequested.load()) {
            resumeRequested.store(true);
            pauseRequested.store(false);
        }
    }
}

void suspendThreads(DWORD processId) {
    HANDLE hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPTHREAD, 0);
#ifdef DEBUG
    if (hSnapshot == INVALID_HANDLE_VALUE) {
        std::cerr << "Failed to create snapshot: " << GetLastError() << std::endl;
        return;
    }

#endif
    THREADENTRY32 te32;
    te32.dwSize = sizeof(THREADENTRY32);
#ifdef DEBUG
    if (!Thread32First(hSnapshot, &te32)) {
        std::cerr << "Failed to retrieve first thread: " << GetLastError() << std::endl;
        CloseHandle(hSnapshot);
        return;
    }
#endif
    do {
        if (te32.th32OwnerProcessID == processId) {
            HANDLE hThread = OpenThread(THREAD_SUSPEND_RESUME, FALSE, te32.th32ThreadID);
#ifdef DEBUG
            if (hThread) {
#endif
                SuspendThread(hThread);
                CloseHandle(hThread);
#ifdef DEBUG
            }
#endif
        }
    } while (Thread32Next(hSnapshot, &te32));
    
    CloseHandle(hSnapshot);
}

// Функція для відновлення потоків процесу
void resumeThreads(DWORD processId) {
    HANDLE hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPTHREAD, 0);
#ifdef DEBUG
    if (hSnapshot == INVALID_HANDLE_VALUE) {
        std::cerr << "Failed to create snapshot: " << GetLastError() << std::endl;
        return;
    }
#endif
    THREADENTRY32 te32;
    te32.dwSize = sizeof(THREADENTRY32);
#ifdef DEBUG
    if (!Thread32First(hSnapshot, &te32)) {
        std::cerr << "Failed to retrieve first thread: " << GetLastError() << std::endl;
        CloseHandle(hSnapshot);
        return;
    }
#endif
    do {
#ifdef DEBUG
        if (te32.th32OwnerProcessID == processId) {
#endif
            HANDLE hThread = OpenThread(THREAD_SUSPEND_RESUME, FALSE, te32.th32ThreadID);
            if (hThread) {
                ResumeThread(hThread);
                CloseHandle(hThread);
#ifdef DEBUG
            }
#endif
        }
    } while (Thread32Next(hSnapshot, &te32));
    
    CloseHandle(hSnapshot);
}

void suspendProcesses(const std::vector<HANDLE>& processes) {
    for (HANDLE hProcess : processes) {
        DWORD processId = GetProcessId(hProcess);
        suspendThreads(processId);
    }
}

void resumeProcesses(const std::vector<HANDLE>& processes) {
    for (HANDLE hProcess : processes) {
        DWORD processId = GetProcessId(hProcess);
        resumeThreads(processId);
    }
}

unsigned int runTest(std::string path, unsigned short numThreads = 4, int prior = NORMAL_PRIORITY_CLASS) {
    std::vector<HANDLE> processes;
    std::ifstream file(path);
    unsigned int digit_count = 0;
    char current_char;
    
    while (file.get(current_char)) {
        if (current_char == '\n') break;
        if (std::isdigit(current_char)) {
            ++digit_count;
            if (file.get(current_char)) {
                if (std::isdigit(current_char)) continue; else file.unget();
            }
        }
    }
    
    if (digit_count < numThreads) {
        std::cerr << "Requested test with " << numThreads << " threads, but matrix got only " << digit_count << " rows! Testing with " << digit_count << " threads." << std::endl;
        numThreads = digit_count;
    }
    
    unsigned int step = digit_count / numThreads;
    std::thread inputThread(listenForPauseResume);  // Запускаємо потік для прослуховування введення
    
    DWORD startTime = GetTickCount();
    for (int i = 0; i < numThreads; ++i) {
        STARTUPINFO si = { sizeof(si) };
        PROCESS_INFORMATION pi;
        std::string cmdline = "worker.exe " + path + ' ' + std::to_string(i * step) + ' ' + std::to_string(i == numThreads - 1 ? (digit_count - step * (numThreads - 1)) : step);
        
        if (CreateProcess(
                NULL,
                const_cast<char*>(cmdline.c_str()),
                NULL,
                NULL,
                FALSE,
                prior,
                NULL,
                NULL,
                &si,
                &pi
                )) {
            processes.push_back(pi.hProcess);
            CloseHandle(pi.hThread); // закриваємо дескриптор потоку
        } else {
            std::cerr << "Failed to create process: " << GetLastError() << std::endl;
        }
    }
    
    bool isSuspended = false;
    
    while (true) {
        DWORD waitResult = WaitForMultipleObjects(processes.size(), processes.data(), TRUE, 10);
        if (waitResult == WAIT_OBJECT_0) break;
        
        std::lock_guard<std::mutex> lock(pauseMutex);
        
        if (pauseRequested.load() && !isSuspended) {
            std::cout << "Processes are being suspended...";  // Повідомлення про призупинення
            suspendProcesses(processes);  // Призупиняємо всі процеси
            isSuspended = true;
            std::cout << " Processes have been suspended. Press Enter to resume..." << std::endl;  // Повідомлення про завершення призупинення
        } else if (resumeRequested.load() && isSuspended) {
            std::cout << "Processes are being resumed...";  // Повідомлення про відновлення
            resumeProcesses(processes);  // Відновлюємо всі процеси
            isSuspended = false;
            resumeRequested.store(false);  // Скидаємо прапорець відновлення
            std::cout << " Processes have been resumed. Press Enter to pause again..." << std::endl;  // Повідомлення про завершення відновлення
        }
    }
    for (HANDLE hProcess : processes) CloseHandle(hProcess);
    inputThread.detach();
    return GetTickCount() - startTime;
}


inline bool exists(const std::string& name) {
    DWORD fileAttributes = GetFileAttributes(name.c_str());
    return (fileAttributes != INVALID_FILE_ATTRIBUTES && !(fileAttributes & FILE_ATTRIBUTE_DIRECTORY));
}
void ExtendConsoleWidth(unsigned short width) {
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    CONSOLE_SCREEN_BUFFER_INFO csbi;
    GetConsoleScreenBufferInfo(hConsole, &csbi);
    if (csbi.dwSize.X>=width)return;
    COORD bufferSize;
    bufferSize.X = width;
    bufferSize.Y = csbi.dwSize.Y;
    SetConsoleScreenBufferSize(hConsole, bufferSize);
    SMALL_RECT windowSize = csbi.srWindow;
    windowSize.Right = windowSize.Left + width - 1;
    SetConsoleWindowInfo(hConsole, TRUE, &windowSize);
}
// void moveCursorToNextLine() {
//     HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
//     CONSOLE_SCREEN_BUFFER_INFO csbi;
//     GetConsoleScreenBufferInfo(hConsole, &csbi);
    
//     COORD newCursorPosition;
//     newCursorPosition.X = 0;
//     newCursorPosition.Y = csbi.dwCursorPosition.Y + 1;
    
//     SetConsoleCursorPosition(hConsole, newCursorPosition);
// }
int main() {
    ExtendConsoleWidth(86);
    std::cout << "--Welcome to os/lab3/os3!--" << std::endl << std::endl;

    std::string filename = "matrix.txt";

    if (!exists(filename)) {
        std::cout << "matrix.txt not found";
        if (exists("genmatrix.exe")) std::cout<<", hovewer it can be generated."<<std::endl<<"1. Generate!"<<std::endl<<"2. Input name"<<std::endl<<"0. Exit"<<std::endl;
        std::cin >> filename;
        while (!exists(filename)){
            if (filename=="1"){
                // std::string N;
                // std::string M;
                // std::cout<<"N? ";
                // std::cin>>N;
                // std::cout<<"M? ";
                // std::cin>>N;
                STARTUPINFO si = { sizeof(si) };
                PROCESS_INFORMATION pi;
            CreateProcess("genmatrix.exe",NULL/*const_cast<char*>(("genmatrix.exe"+N+" "+M).c_str())*/,NULL,NULL,0,0,NULL,NULL,&si,&pi);
            WaitForSingleObject(pi.hProcess, INFINITE);
            filename="matrix.txt";}
            else if (filename=="0")return 0;else if (filename=="2")continue;else {std::cout<<"Wrong input."<<std::endl;std::cin>>filename;}}getchar();
    } else std::cout << "File matrix.txt found." << std::endl;

    int prior = NORMAL_PRIORITY_CLASS;

    char choice;
    while (true) {
        std::cout << "1. Run tests" << std::endl;
        std::cout << "2. Set priority" << std::endl;
        std::cout << "3. Change attributes..." << std::endl;
        std::cout << "0. Exit" << std::endl;
        choice = std::getchar();
        if (choice == '1') {
            std::getchar();
            for (int i = 1; i <= std::thread::hardware_concurrency()*2; ++i) {
                std::string line=std::to_string(i)+" threads: "+std::to_string(runTest(filename, i, prior))+"ms\n";
                WriteFile(outfile,line.c_str(),line.size(),0,0);
            }break;
        } else if (choice == '2') {
            std::cout << "Enter new value for priority:" << std::endl
                      << "  64 for IDLE_PRIORITY_CLASS" << std::endl
                      << "  16384 for BELOW_NORMAL_PRIORITY_CLASS" << std::endl
                      << "  32 for NORMAL_PRIORITY_CLASS" << std::endl
                      << "  32768 for ABOVE_NORMAL_PRIORITY_CLASS" << std::endl
                      << "  128 for HIGH_PRIORITY_CLASS" << std::endl
                      << "  256 for REALTIME_PRIORITY_CLASS" << std::endl
                      << "Your choice: ";
            std::cin >> prior;
            std::cin.ignore(std::numeric_limits<std::streamsize>::max(), '\n'); // Clear input buffer
        } else if (choice == '3'){
            std::cout<<"Choose attrib..."<<std::endl<<"1. FILE_ATTRIBUTE_NORMAL"<<std::endl<<"2. FILE_ATTRIBUTE_HIDDEN"<<std::endl
                    <<"3. FILE_ATTRIBUTE_READONLY"<<std::endl<<"4. FILE_ATTRIBUTE_SYSTEM"<<std::endl<<"5. FILE_ATTRIBUTE_ARCHIVE"<<std::endl
                    <<"6. FILE_ATTRIBUTE_TEMPORARY"<<std::endl<<"7. FILE_ATTRIBUTE_COMPRESSED"<<std::endl<<"8. FILE_ATTRIBUTE_ENCRYPTED"<<std::endl
                    <<"9. FILE_ATTRIBUTE_DIRECTORY"<<std::endl;
            short attribs;
            std::cin>>attribs;
            if (attribs==1)attribs=FILE_ATTRIBUTE_NORMAL;
            else if (attribs==2)attribs=FILE_ATTRIBUTE_HIDDEN;
            else if (attribs==3)attribs=FILE_ATTRIBUTE_READONLY;
            else if (attribs==4)attribs=FILE_ATTRIBUTE_SYSTEM;
            else if (attribs==5)attribs=FILE_ATTRIBUTE_ARCHIVE;
            else if (attribs==6)attribs=FILE_ATTRIBUTE_TEMPORARY;
            else if (attribs==7)attribs=FILE_ATTRIBUTE_COMPRESSED;
            else if (attribs==8)attribs=FILE_ATTRIBUTE_ENCRYPTED;
            else if (attribs==9)attribs=FILE_ATTRIBUTE_DIRECTORY;
            else {std::cout<<"Wrong input."<<std::endl;continue;}
            SetFileAttributes("out",attribs);
            getchar();
          }
        else if (choice == '0')break;
        else std::cout << "Wrong input." << std::endl;
    }
    return 0;
}
